#!/usr/bin/env python
import numpy as np
import glfw
from gym import Env
from gym.spaces import Box
from mujoco_py import MjSim, MjViewer, const

from gym_morph import SPECIES, TASK
from gym_morph.model import build_robot


class MorphEnv(Env):
    metadata = {'render.modes': ['human']}

    def __init__(self, species='ant', task='run'):
        assert species in SPECIES, 'Invalid species "{}"'.format(species)
        assert task in TASK, 'Invalid task "{}"'.format(task)
        self.species = species
        self.task = task
        self.current_seed = 0  # Used to generate random initial states
        self.sim = MjSim(build_robot(species))
        ctrlrange = self.sim.model.actuator_ctrlrange
        self.action_space = Box(ctrlrange[:, 0], ctrlrange[:, 1])
        self.viewer = None

    def _reset(self):
        self.current_seed += 1
        rs = np.random.RandomState(self.current_seed)

        for i, q in enumerate(self.sim.model.jnt_qposadr):
            if self.sim.model.jnt_type[i] != const.JNT_FREE:
                jnt_range = self.sim.model.jnt_range[i]
                self.sim.data.qpos[q] = rs.uniform(jnt_range[0], jnt_range[1])
            else:  # Free joint, only set angle, not position
                quat = rs.uniform(0, 1, size=4)  # Random quaternion
                quat /= np.sqrt(np.sum(np.square(quat)))  # Normalize
                self.sim.data.qpos[q + 3:q + 7] = quat

        # TODO: Use seed to generate random goal
        self.last_torso_xpos = self.sim.data.get_body_xpos('torso').copy()
        return self._get_obs()

    def _get_obs(self):
        # TODO: input goal locations relative to body pose
        return np.concatenate([self.sim.data.qpos, self.sim.data.qvel])

    def _get_reward(self):
        if self.task == 'run':
            # Reward is velocity in the x direction
            xpos = self.sim.data.get_body_xpos('torso')
            reward = xpos[0] - self.last_torso_xpos[0]
            self.last_torso_xpos = xpos.copy()
            return reward
        return 0.

    def _set_action(self, action):
        action = np.asarray(action)
        assert self.action_space.contains(action), '{}'.format(action)
        self.sim.data.ctrl[:] = action

    def _step(self, action):
        self._set_action(action)
        self.sim.step()
        obs = self._get_obs()
        reward = self._get_reward()
        done = False  # TODO
        info = {}  # TODO
        return obs, reward, done, info

    def _render(self, mode='human', close=False):
        assert mode == 'human'  # TODO: rgb_array rendering
        if close and self.viewer is not None:
            glfw.destroy_window(self.viewer.window)
        else:
            if self.viewer is None:
                self.viewer = MjViewer(self.sim)
            self.viewer.render()

    def _seed(self, seed=None):
        if seed is not None:
            self.seed = seed
        return [self.seed]



#!/usr/bin/env python
# model.py - build the species and task models

import numpy as np
from mujoco_py import load_model_from_xml

from gym_morph import SPECIES

XML = '''
<mujoco>
  <compiler angle="degree" coordinate="local" inertiafromgeom="true"/>
  <option integrator="RK4" timestep="0.01"/>
  <default>
    <joint armature="1" damping="1" limited="true"/>
    <geom conaffinity="0" condim="3" density="5.0" friction="1 0.5 0.5" margin="0.01" rgba="0.8 0.6 0.4 1" type="capsule"/>
    <motor ctrllimited="true" ctrlrange="-1.0 1.0" gear="150"/>
  </default>
  <worldbody>
    <site name="origin" pos="0 0 0" rgba="1 0 0 1" size=".1"/>
    <geom name="floor" conaffinity="1" condim="3" pos="0 0 0" rgba="0.8 0.9 0.8 1" size="40 40 40" type="plane"/>
    <body name="torso" pos="0 0 1">
      <geom size="0.25" type="sphere"/>
      <joint armature="0" damping="0" limited="false" margin="0.01" name="root" pos="0 0 0" type="free"/>
      {legs}
    </body>
  </worldbody>
  <actuator>
    {motors}
  </actuator>
</mujoco>
'''

LEG = '''
  <body name="{name}" pos="0 0 0">
    <geom fromto="0.0 0.0 0.0 {pos}" size="0.08"/>
    <body pos="{pos}">
      <joint axis="0 0 1" name="{name}_j1" pos="0.0 0.0 0.0" range="-30 30" type="hinge"/>
      <geom fromto="0.0 0.0 0.0 {pos}" size="0.08"/>
      <body pos="{pos}">
        <joint axis="{axis}" name="{name}_j2" pos="0.0 0.0 0.0" range="30 70" type="hinge"/>
        <geom fromto="0.0 0.0 0.0 {pos2}" size="0.08"/>
      </body>
    </body>
  </body>
'''
MOTOR = '''
  <motor joint="{name}_j1"/>
  <motor joint="{name}_j2"/>
'''


def vec2str(v):
    return ' '.join(str(i) for i in v)


def build_robot(species):
    ''' Build and return a MuJoCo model for a given species. '''
    assert species in SPECIES
    if species == 'ant':
        legs = ''
        motors = ''
        for i, angle in enumerate(np.arange(0, np.pi * 2, np.pi / 2)):
            pos = np.array([np.sin(angle), np.cos(angle), 0]) * .28284
            axis = np.cross([0, 0, 1], pos)
            name = 'leg_{}'.format(i)
            legs += LEG.format(name=name, pos=vec2str(pos),
                               pos2=vec2str(2 * pos), axis=vec2str(axis))
            motors += MOTOR.format(name=name)
        xml = XML.format(legs=legs, motors=motors)
        return load_model_from_xml(xml)



#!/usr/bin/env python

from setuptools import find_packages, setup


setup(name='gym-morph',
      version='0.0.2',
      description='Gym mujoco environments with morphology-control',
      packages=find_packages(),
      author='Alex Ray',
      author_email='a@machinaut.com',
      install_requires=['numpy>=1.13.3',
                        'gym>=0.9.4',
                        'glfw>=1.4.0',
                        'mujoco_py>=1.50.1.23'],
      include_package_data=True)



from gym.envs.registration import register

# Based on the MuJoCo Ant model
SPECIES = ['longleg',  # morphing - extends first leg joint
           'longfoot',  # morphing - extends second leg joint
           'insect',  # non-morphing - 6 two-joint legs
           'lifter',  # non-morphing - extra lifing joint in hip
           'triple',  # non-morphing - three leg joints
           'ant']  # non-morphing - 4 two-joint legs

TASK = ['run',  # Run as fast as possible in a single direction
        'flagrun',  # Run to a specified 2d location
        'twister']  # Place foot on a specified spot


for species in SPECIES:
    for task in TASK:
        name = '-'.join(['Morph', species, task, 'v0'])
        kwargs = dict(species=species, task=task)
        register(id=name, entry_point='gym_morph.env:MorphEnv', kwargs=kwargs,
                 max_episode_steps=1000)
